// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if 0
#define DEBUG_PRINT
#endif

#include "actgen_cooperative_offside_line_pass.h"
#include "simple_pass_checker.h"
#include "cooperative_action.h"
#include "action_state_pair.h"
#include "pass.h"

#include <rcsc/common/logger.h>
#include <rcsc/math_util.h>

#include <set>
#include <utility>
#ifdef DEBUG_PRINT
#include <iostream>
#endif

ActGen_CooperativeOffsideLinePass::ActGen_CooperativeOffsideLinePass()
{
}

ActGen_CooperativeOffsideLinePass::~ActGen_CooperativeOffsideLinePass()
{
}


static
double
s_get_ball_speed_for_pass( const double & distance );

void
ActGen_CooperativeOffsideLinePass::generate( std::vector< ActionStatePair > * result,
                                             const rcsc::PredictState & state,
                                             const rcsc::WorldModel &,
                                             const std::vector< ActionStatePair > & path ) const
{
    static const int VALID_PLAYER_THRESHOLD = 8;

#ifdef DEBUG_PRINT
    rcsc::dlog.addText( rcsc::Logger::ACTION,
                        __FILE__": ActGen_CooperativeOffsideLinePass::addCandidates()" );
#endif

    if ( path.size() == 0 )
    {
        return;
    }

    std::set< double > opp_y_set;

    const rcsc::PlayerCont::const_iterator opp_end = state.opponents().end();
    for ( rcsc::PlayerCont::const_iterator opp = state.opponents().begin();
          opp != opp_end;
          ++opp )
    {
        if ( (*opp).posCount() > VALID_PLAYER_THRESHOLD
             || (*opp).isGhost() )
        {
            continue;
        }

        if ( ! (*opp).pos().x < state.offsideLineX() - 10.0 )
        {
            continue;
        }

        opp_y_set.insert( (*opp).pos().y );
    }
    opp_y_set.insert( - rcsc::ServerParam::i().pitchHalfWidth() );
    opp_y_set.insert( + rcsc::ServerParam::i().pitchHalfWidth() );

    std::vector< double > y_candidates;
    {
        const std::set< double >::const_iterator opp_y_set_end = opp_y_set.end();
        const std::set< double >::const_iterator opp_y_set_begin = opp_y_set.begin();
        double old_y = 0.0;
        for ( std::set< double >::const_iterator opp_y = opp_y_set.begin();
              opp_y != opp_y_set_end;
              ++opp_y )
        {
            if ( opp_y != opp_y_set_begin )
            {
                y_candidates.push_back( ( *opp_y + old_y ) / 2.0 );
            }

            old_y = *opp_y;
        }
    }


    std::vector< double > length_candidates;

    for ( double d = -2.0; d <= 30.0 + rcsc::EPS; )
    {
        length_candidates.push_back( d );

        if ( path.size() >= 1 )
        {
            d += 10.0;
        }
        else if ( d >= 5.0 + rcsc::EPS )
        {
            d += 2.0;
        }
        else if ( d >= 10.0 + rcsc::EPS )
        {
            d += 5.0;
        }
        else
        {
            d += 1.0;
        }
    }

    const rcsc::AbstractPlayerObject * holder = state.ballHolder();
    const rcsc::Vector2D ball_pos = state.ball().pos();

    //
    // add pass candidates
    //
    const SimplePassChecker pass_check;
    const unsigned long spendTime = state.spendTime();

    const PredictPlayerPtrCont::const_iterator end = state.allTeammates().end();
    for ( PredictPlayerPtrCont::const_iterator pl = state.allTeammates().begin();
          pl != end;
          ++pl )
    {
        if ( ! (*pl)->isValid()
             || (*pl)->posCount() > VALID_PLAYER_THRESHOLD
             || (*pl)->isGhost()
             || (*pl)->unum() == holder->unum()
             || (*pl)->unum() == -1 )
        {
            continue;
        }

#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::ACTION,
                            __FILE__": checking offside line pass to player %d",
                            (*pl)->unum() );
#endif

        for ( size_t y_index = 0; y_index < y_candidates.size(); ++y_index )
        {
            const rcsc::Vector2D receive_point( state.offsideLineX(),
                                                y_candidates[y_index] );

            const double ball_dist = ( state.ball().pos() - receive_point ).r();
            const double receiver_dist = ( (*pl)->pos() - receive_point ).r();
            const double ball_first_speed = s_get_ball_speed_for_pass( ball_dist );

            const long ball_step = rcsc::calc_length_geom_series( ball_first_speed,
                                                                  ball_dist,
                                                                  rcsc::ServerParam::i().ballDecay() );

            if ( (*pl)->playerTypePtr()->cyclesToReachDistance( receiver_dist )
                 > static_cast< double >( spendTime ) + ball_step )
            {
#ifdef DEBUG_PRINT
                rcsc::dlog.addText( rcsc::Logger::ACTION,
                                    __FILE__": cannot player %d can reach to on offside line %f",
                                    (*pl)->unum(), receive_point.y );
#endif
                continue;
            }

            rcsc::dlog.addText( rcsc::Logger::ACTION,
                                __FILE__": player %d can reach to on offside line %f",
                                (*pl)->unum(), receive_point.y );

            if ( ! pass_check( state, *holder, **pl, receive_point, ball_first_speed ) )
            {
                continue;
            }

            rcsc::PredictState::ConstPtr result_state( new rcsc::PredictState( state,
                                                                               ball_step,
                                                                               (*pl)->unum(),
                                                                               receive_point ) );

#if 0
            CooperativeAction::Ptr action( new rcsc::CooperativeAction( rcsc::CooperativeAction::Pass,
                                                                              "cooperative_offside_line" ) );
            action->setTargetPoint( receive_point );
            action->setTargetPlayerUnum( (*pl)->unum() );
            action->setBallSpeed( ball_first_speed );
            //action->setReceiverReachStep();
            //action->setBallReachStep();
            action->setFinalAction( true );

            action->addSupportPlayerMove( rcsc::CooperativeAction::SupportPlayerMove( (*pl)->unum(),
                                                                                      receive_point ) );
#else

            rcsc::CooperativeAction::Ptr action( new rcsc::Pass( holder->unum(),
                                                                 (*pl)->unum(),
                                                                 (*pl)->pos(),
                                                                 ball_first_speed,
                                                                 ball_step,
                                                                 2,
                                                                 false,
                                                                 "CooperativeOffsideLinePass" ) );
#endif

            result->push_back( ActionStatePair( action, result_state ) );

#ifdef DEBUG_PRINT
            std::cerr << state.time() << "cooperative offside line pass"
                      << std::endl;
#endif
        }
    }
}


static
double
s_get_ball_speed_for_pass( const double & distance )
{
    if ( distance >= 20.0 )
    {
        return 3.0;
    }
    else if ( distance >= 8.0 )
    {
        return 2.5;
    }
    else if ( distance >= 5.0 )
    {
        return 1.8;
    }
    else
    {
        return 1.5;
    }
}
