// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "local_positioning.h"
#include <rcsc/common/server_param.h>
#include <rcsc/player/player_predicate.h>
#include <rcsc/player/player_evaluator.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/geom/sector_2d.h>
#include <algorithm>
#include <cmath>
#include <cfloat>

#include "puppets/param_field_evaluator.h"
#include "puppets/puppets_player.h"
#include "chain_action/action_chain_graph.h"
#include "chain_action/predict_state.h"
#include "chain_action/field_evaluator.h"
#include "chain_action/action_generator.h"
#include "chain_action/simple_pass_checker.h"
#include "field_analyzer.h"
#include "strategy.h"

#if 0
#define NEW_LOCAL_POSITIONING 1
#endif

#if 1
#define OFFENSE_ONLY
#endif


#if 0
static const double  POSITION_DEFENCE_PASS_CUT_RATE       = 0.985763;
static const double  POSITION_FREE_FORWARD_WIDTH_RATE     = 0.112534;
static const double  POSITION_MARK_COUNT_RATE             = 0.302176;
static const double  POSITION_OFFENCE_PASS_CUT_RATE       = 0.424277;
static const double  POSITION_OPPONENT_FREE_RADIUS_RATE   = 0.591179;
static const double  POSITION_PASS_COURSE_FREE_ANGLE_RATE = 0.436559;
static const double  POSITION_SHOOT_WIDTH_RATE            = 0.110574;
static const double  POSITION_TEAMMATE_FREE_RADIUS_RATE   = 0.142917;
#else
static const double  POSITION_DEFENCE_PASS_CUT_RATE       = 0.0;
static const double  POSITION_FREE_FORWARD_WIDTH_RATE     = 0.112534;
static const double  POSITION_MARK_COUNT_RATE             = 0.0;
static const double  POSITION_OFFENCE_PASS_CUT_RATE       = 0.424277;
static const double  POSITION_OPPONENT_FREE_RADIUS_RATE   = 0.591179;
static const double  POSITION_PASS_COURSE_FREE_ANGLE_RATE = 0.436559;
static const double  POSITION_SHOOT_WIDTH_RATE            = 0.110574;
static const double  POSITION_TEAMMATE_FREE_RADIUS_RATE   = 0.0;
#endif


static
double
evaluatePoint( const rcsc::WorldModel & wm,
               const rcsc::Vector2D & point,
               bool have_advantage,
               const rcsc::Vector2D & ball_pos,
               double offside_buffer );

static
double
get_offside_buffer( const rcsc::WorldModel & wm,
                    const rcsc::Vector2D & ball_pos,
                    long our_team_access_step,
                    long opponent_team_access_step );

#if 0
static
double
get_teammate_free_radius( const rcsc::WorldModel & wm,
                          const rcsc::Vector2D & point );
#endif

static
double
get_opponent_free_radius( const rcsc::WorldModel & wm,
                          const rcsc::Vector2D & point );

static
double
get_free_course_angle( const rcsc::WorldModel & wm,
                       const rcsc::Vector2D & from,
                       const rcsc::Vector2D & to );

static
double
get_free_forward_width( const rcsc::WorldModel & wm,
                        const rcsc::Vector2D & point );

#if 0
static
int
get_pass_cut_count( const rcsc::WorldModel & wm,
                    const rcsc::Vector2D & point,
                    const rcsc::Vector2D & ball_pos );
#endif

#if 0
static
int
get_mark_count( const rcsc::WorldModel & wm,
                const rcsc::Vector2D & point,
                const rcsc::Vector2D & ball_pos );
#endif


static
double getDistPlayerNearestToPoint( const rcsc::WorldModel &,
                                    const rcsc::Vector2D & point,
                                    const rcsc::PlayerCont & cont,
                                    int validOpponentThreshold = -1 );

static
double getDistOpponentNearestToPoint( const rcsc::WorldModel & wm,
                                      const rcsc::Vector2D & point,
                                      int validOpponentThreshold = -1 );

rcsc::Vector2D
LocalPositioning::getLocalAdjustmentPoint( const rcsc::PlayerAgent * agent,
                                           const rcsc::Vector2D & base_point )
    const
{
    const rcsc::WorldModel & wm = agent->world();

    const int teammate_reach_cycle = wm.interceptTable()->teammateReachCycle();
    const int opponent_reach_cycle = wm.interceptTable()->opponentReachCycle();

    const bool have_advantage = (teammate_reach_cycle < opponent_reach_cycle);

    const int ball_predict_cycle = std::min( teammate_reach_cycle,
                                             opponent_reach_cycle);

    const rcsc::Vector2D ball_predict_point = wm.ball().inertiaPoint( ball_predict_cycle );

    const double offside_buffer = get_offside_buffer( wm,
                                                      ball_predict_point,
                                                      teammate_reach_cycle,
                                                      opponent_reach_cycle );
#ifdef OFFENSE_ONLY
    if ( ! have_advantage )
    {
        return base_point;
    }

    double defense_line_x = wm.ourDefenseLineX();
    double x_width = wm.offsideLineX() - defense_line_x; 

#if 0
    if ( base_point.x < (defense_line_x + x_width * 0.2)
         && base_point.x < 0.0 )
#elif 1
    if ( base_point.x < (defense_line_x + x_width * 0.2) )
#elif 0
    if ( base_point.x < (defense_line_x + x_width * 0.9) )
#else
    if ( base_point.x < (defense_line_x + x_width * 0.9)
         && base_point.x < 20.0 )
#endif
    {
        return base_point;
    }
#endif


    std::vector<double> searchRadiusList;
#if 0
# if NEW_LOCAL_POSITIONING
    searchRadiusList.push_back( 0.50 );
    searchRadiusList.push_back( 1.00 );
    searchRadiusList.push_back( 1.50 );
    searchRadiusList.push_back( 2.00 );
# else
    searchRadiusList.push_back( 1.0 );
    searchRadiusList.push_back( 2.0 );
    searchRadiusList.push_back( 3.0 );
    searchRadiusList.push_back( 5.0 );
# endif
#elif 0
    searchRadiusList.push_back(  2.5 );
    searchRadiusList.push_back(  5.0 );
    searchRadiusList.push_back(  7.5 );
    searchRadiusList.push_back( 10.0 );
#else
    searchRadiusList.push_back(  2.0 );
    searchRadiusList.push_back(  3.0 );
    searchRadiusList.push_back(  5.0 );
#endif

    rcsc::Vector2D best_position = base_point;
    double maxEvaluation = evaluatePoint( wm,
                                          best_position,
                                          have_advantage,
                                          ball_predict_point,
                                          offside_buffer );


#if NEW_LOCAL_POSITIONING
    const int N_DIVISION = 16;
#else
    const int N_DIVISION = 8;
#endif

    for ( size_t r = 0; r < searchRadiusList.size(); ++r )
    {
        rcsc::AngleDeg angleOffset = ((r % 2 == 0) ?
                                      rcsc::AngleDeg(0.0)
                                      : rcsc::AngleDeg(360.0 / N_DIVISION / 2));

        for ( int i = 0; i < N_DIVISION; ++i )
        {
            rcsc::Vector2D p = base_point;

            p += rcsc::Vector2D::polar2vector( searchRadiusList[r],
                                               rcsc::AngleDeg( i * 360.0 / N_DIVISION )
                                               + rcsc::AngleDeg( angleOffset ) );

            double ev = evaluatePoint( wm,
                                       p,
                                       have_advantage,
                                       ball_predict_point,
                                       offside_buffer );

            if ( ev > maxEvaluation )
            {
                maxEvaluation = ev;
                best_position = p;
            }
        }
    }

    return best_position;
}

#if 0
static
double
get_offside_buffer( const rcsc::WorldModel & wm,
                    const rcsc::Vector2D & ball_pos,
                    long  our_team_access_step,
                    long  opponent_team_access_step )
{
    if ( wm.offsideLineX() - 0.01 < ball_pos.x
         && our_team_access_step < opponent_team_access_step
         && our_team_access_step <= 3 )
    {
        return 1.0;
    }
    else if ( opponent_team_access_step - our_team_access_step >= 3 )
    {
        return 2.0;
    }
    else
    {
        return 3.0;
    }
}
#else
static
double
get_offside_buffer( const rcsc::WorldModel &,
                    const rcsc::Vector2D &,
                    long,
                    long )
{
    return 0.5;
}
#endif

#if 0
static
double
evaluatePoint( const rcsc::WorldModel & wm,
               const rcsc::Vector2D & point,
               bool have_advantage,
               const rcsc::Vector2D & ball_pos,
               double offside_buffer )
{
    // out of field
    if ( point.x < rcsc::ServerParam::i().ourTeamGoalLineX() - 2.0
         || point.x > rcsc::ServerParam::i().theirTeamGoalLineX() - 2.0
         || point.absY() >= rcsc::ServerParam::i().pitchHalfWidth() - 2.0 )
    {
        return 0.0;
    }


    static_cast<void>( have_advantage );
    static_cast<void>( ball_pos );
    static_cast<void>( offside_buffer );

    const double opponent_free_radius = get_opponent_free_radius( wm, point );
    const double pass_course_free_angle = get_free_course_angle( wm, point, ball_pos );

    double evaluation = 0.0;

    //
    // add bonus for opponent free radius
    //
    if ( have_advantage )
    {
        evaluation += std::min( opponent_free_radius, 10.0 );
    }

    //
    // add bonus for pass course free angle
    //
    if ( have_advantage )
    {
        if ( pass_course_free_angle >= 15.0
             && ( point - ball_pos ).r() < 30.0 )
        {
            evaluation += 20.0;
        }
    }

    return evaluation;
}
#endif



#if 0
static
double
evaluatePointOriginal( const rcsc::WorldModel & wm,
                       const rcsc::Vector2D & point,
                       bool have_advantage,
                       const rcsc::Vector2D & ball_pos,
                       double offside_buffer )
{
    //
    // check out of field
    //
    if ( point.x < rcsc::ServerParam::i().ourTeamGoalLineX() - 2.0
         || point.x > rcsc::ServerParam::i().theirTeamGoalLineX() - 2.0
         || point.absY() >= rcsc::ServerParam::i().pitchHalfWidth() - 2.0 )
    {
        return 0.0;
    }

    if ( point.x < wm.ourDefenseLineX()
         || point.x >= wm.offsideLineX() - offside_buffer )
    {
        return 0.0;
    }


    //
    // free radius
    //
    const double teammate_free_radius = get_teammate_free_radius( wm, point );

    double evaluation = 0.0;

    if ( have_advantage )
    {
        const double opponent_free_radius
            = get_opponent_free_radius( wm, point );

        evaluation = teammate_free_radius
                     * POSITION_TEAMMATE_FREE_RADIUS_RATE * 100.0
                     + opponent_free_radius
                     * POSITION_OPPONENT_FREE_RADIUS_RATE * 100.0;
    }
    else
    {
        evaluation = teammate_free_radius;
    }


#if 0
    //
    // dist from opponent goal
    //
    if ( have_advantage
         && (point - rcsc_ext::ServerParam_opponentTeamGoalPos()).r() < 35.0 )
    {
        double shoot_free_angle = rcsc::Bhv_Shoot::free_angle( f );

        if ( shoot_free_angle >= 15.0 )
        {
            evaluation += (100000.0
                           + shoot_free_angle
                           * 1000.0
                           * POSITION_SHOOT_WIDTH_RATE );
        }
    }
#endif

    //
    // pass course
    //
    if ( have_advantage )
    {
        double pass_course_free_angle = get_free_course_angle( wm, point, ball_pos );

        if ( pass_course_free_angle >= 10.0 )
        {
            evaluation += 5000.0;

            evaluation += std::min( (point - ball_pos).r(), 5.0 ) * 1000.0;
        }

        if ( pass_course_free_angle >= 15.0 )
        {
            evaluation += 5000.0;
        }

        evaluation += std::min( pass_course_free_angle, 45.0 )
                      * 100.0
                      * POSITION_PASS_COURSE_FREE_ANGLE_RATE;
    }

    //
    // forward near offside line
    //
    if ( have_advantage
         && wm.getPlayerCont( new rcsc::AndPlayerPredicate
                              ( new rcsc::TeammatePlayerPredicate( wm ),
                                new rcsc::CoordinateAccuratePlayerPredicate( 6 ),
                                new rcsc::XCoordinateForwardPlayerPredicate
                                ( wm.offsideLineX() - 10.0 ) ) ).size() <= 3 )
    {
        double free_forward_width = get_free_forward_width( wm, point );

        double line_dist = wm.offsideLineX() - point.x;

        if ( line_dist < - 10.0 )
        {
            if ( free_forward_width >= 5.0 )
            {
                evaluation += 50000.0;
            }

            evaluation += free_forward_width * (10.0 - line_dist)
                          * 15000.0
                          * POSITION_FREE_FORWARD_WIDTH_RATE;

            evaluation += free_forward_width * (10.0 - line_dist)
                          * 150000.0
                          * std::max( 0.0, 10.0 - teammate_free_radius );
        }
    }

    //
    // pass cut
    //
    if ( have_advantage )
    {
        evaluation += 1000.0 * POSITION_OFFENCE_PASS_CUT_RATE
                      * get_pass_cut_count( wm, point, ball_pos );
    }
    else
    {
        evaluation += 10000.0 * POSITION_DEFENCE_PASS_CUT_RATE
                      * get_pass_cut_count( wm, point, ball_pos );
    }

    //
    // mark count
    //
    if ( ! have_advantage )
    {
        evaluation += 100000.0 * POSITION_MARK_COUNT_RATE
                      * get_mark_count( wm, point, ball_pos );
    }


#if 0
    rcsc::Vector2D opposite_goal_post = rcsc::Vector2D( wm.our_team().goal_line_x(),
                                                        wm.ball().coordinate().y_sign()
                                                        * wm.sserver_param().goal_width() );

    if ( ! have_advantage
         && (point - opposite_goal_post).r() < 5.0
         && wm.ball().coordinate().x() < wm.our_team().penalty_area_line_x()
         && wm.ball().coordinate().y() * point.y() < 0.0
         && wm.getPlayerCont( new And_Player_Predicate
                              ( new Our_Team_Accurate_Field_Player_Predicate( -20 ),
                                new Point_Close_Player_Predicate
                                ( wm,
                                  rcsc::Vector2D( wm.our_team().goal_line_x(),
                                                  wm.ball().coordinate().y_sign()
                                                  * wm.sserver_param().goal_width() ),
                                  5.0 ) ) ).empty() )
    {
        evaluation += 500000.0;
    }
#endif

    return evaluation;
}
#endif


static
double
evaluatePointDevel( const rcsc::WorldModel & wm,
                    const rcsc::Vector2D & point,
                    bool have_advantage,
                    const rcsc::Vector2D & ball_pos,
                    double offside_buffer )
{
    static const int VALID_PLAYER_ACCURACY = 20;

    //
    // check out of field
    //
    if ( point.x < rcsc::ServerParam::i().ourTeamGoalLineX() - 2.0
         || point.x > rcsc::ServerParam::i().theirTeamGoalLineX() - 2.0
         || point.absY() >= rcsc::ServerParam::i().pitchHalfWidth() - 2.0 )
    {
        return 0.0;
    }

    if ( point.x < wm.ourDefenseLineX()
         || point.x >= wm.offsideLineX() - offside_buffer )
    {
        return 0.0;
    }


    //
    // free radius
    //
    double evaluation = 0.0;

    if ( have_advantage )
    {
        const double opponent_free_radius
            = get_opponent_free_radius( wm, point );

        evaluation += opponent_free_radius * 3.0;
    }


    //
    // pass course
    //
    if ( have_advantage )
    {
        double pass_course_free_angle = get_free_course_angle( wm, point, ball_pos );

        if ( pass_course_free_angle >= 10.0 )
        {
            evaluation += 100.0;

            evaluation += std::min( pass_course_free_angle, 30.0 ) * 5.0;
        }
    }


    //
    // forward near offside line
    //
    if ( have_advantage )
    {
        double free_forward_width = get_free_forward_width( wm, point );

        double line_dist = wm.offsideLineX() - point.x;

        if ( line_dist < 10.0 )
        {
            if ( free_forward_width >= 5.0 )
            {
                evaluation += 10.0;
            }

            evaluation += std::min( free_forward_width, 10.0 )
		                * (10.0 - line_dist) * 10.0;
        }
    }


    //
    // pass_count
    //
    const rcsc::PredictState state( wm );
    const SimplePassChecker pass_checker;
    const int pass_count = FieldAnalyzer::get_pass_count( state,
                                                          pass_checker,
                                                          2.5,
                                                          3 );

    const double pass_count_bonus = std::min( pass_count, 3 ) * 50.0;
    evaluation += pass_count_bonus;


    if ( FieldAnalyzer::can_shoot_from
         ( true,
           point,
           wm.getPlayerCont( new rcsc::OpponentOrUnknownPlayerPredicate( wm.ourSide() ) ),
           VALID_PLAYER_ACCURACY ) )
    {
        evaluation += 100.0;
    }

    return evaluation;
}



#if 0
static
double
evaluatePoint( const rcsc::WorldModel & wm,
               const rcsc::Vector2D & point,
               bool have_advantage,
               const rcsc::Vector2D & ball_pos,
               double offside_buffer )
{
    return evaluatePointOriginal( wm, point, have_advantage,
                                  ball_pos, offside_buffer );
}
#elif 0
static
double
evaluatePoint( const rcsc::WorldModel & wm,
               const rcsc::Vector2D & point,
               bool have_advantage,
               const rcsc::Vector2D & ball_pos,
               double offside_buffer )
{
    return evaluatePointDevel( wm, point, have_advantage,
                               ball_pos, offside_buffer );
}
#elif 1
static
double
evaluatePoint( const rcsc::WorldModel & wm,
               const rcsc::Vector2D & point,
               bool have_advantage,
               const rcsc::Vector2D & ball_pos,
               double offside_buffer )
{
    if ( ! have_advantage )
    {
        return evaluatePointDevel( wm, point, have_advantage,
                                   ball_pos, offside_buffer );
    }


    if ( point.x > wm.offsideLineX() - offside_buffer )
    {
        return - DBL_MAX / 2.0;
    }

    ParamFieldEvaluator evaluator;
    const rcsc::PredictState s( wm );

    // when pass received
    const rcsc::PredictState state( s, 
                                    0, // XXX: should change to time by ball dist
                                    wm.self().unum(),
                                    point );

    return evaluator( state, std::vector< ActionStatePair >() );
}
#elif 0
static
double
evaluatePoint( const rcsc::WorldModel & wm,
               const rcsc::Vector2D & point,
               bool have_advantage,
               const rcsc::Vector2D & ball_pos,
               double offside_buffer )
{
    if ( ! have_advantage )
    {
        return evaluatePointOriginal( wm, point, have_advantage,
                                      ball_pos, offside_buffer );
    }
}
#else
static
double
evaluatePoint( const rcsc::WorldModel & wm,
               const rcsc::Vector2D & point,
               bool have_advantage,
               const rcsc::Vector2D & ball_pos,
               double offside_buffer )
{
    if ( ! have_advantage )
    {
        return evaluatePointOriginal( wm, point, have_advantage,
                                      ball_pos, offside_buffer );
    }

    const size_t EVALUATE_LIMIT = ActionChainGraph::DEFAULT_MAX_EVALUATE_LIMIT / 10;

    FieldEvaluator::ConstPtr evaluator = PuppetsPlayer::s_create_field_evaluator();
    ActionGenerator::ConstPtr generator = PuppetsPlayer::s_create_action_generator();

    // XXX: not implemented yet
    const ActionChainGraph::ConstPtr
        g = ActionChainGraph( evaluator, generator, wm, state,
              rcsc::RecursiveActionChainGraph::DEFAULT_MAX_CHAIN_LENGTH,
              EVALUATE_LIMIT );

    return g->getBestEvaluation();
}
#endif


#if 0
static
double
get_teammate_free_radius( const rcsc::WorldModel & wm,
                          const rcsc::Vector2D & point )
{
    static const int VALID_PLAYER_ACCURACY = 20;

    return getDistOpponentNearestToPoint( wm, point, VALID_PLAYER_ACCURACY );
}
#endif

static
double
get_opponent_free_radius( const rcsc::WorldModel & wm,
                          const rcsc::Vector2D & point )
{
    static const int VALID_PLAYER_ACCURACY = 20;

    return getDistOpponentNearestToPoint( wm, point, VALID_PLAYER_ACCURACY );
}

static
double
get_free_course_angle( const rcsc::WorldModel & wm,
                       const rcsc::Vector2D & from,
                       const rcsc::Vector2D & to )
{
    const int VALID_PLAYER_ACCURACY = 40;

    rcsc::AbstractPlayerCont opp_set
        = wm.getPlayerCont( new rcsc::AndPlayerPredicate
                            ( new rcsc::OpponentOrUnknownPlayerPredicate( wm ),
                              new rcsc::CoordinateAccuratePlayerPredicate
                              ( VALID_PLAYER_ACCURACY ) ) );

    double free_course_angle;
    free_course_angle = std::min( 180.0,
                                  rcsc::AbstractPlayerObject::get_minimum_evaluation
                                  ( opp_set,
                                    new rcsc::AbsAngleDiffPlayerEvaluator
                                    ( from, (to - from).th() ) ) );

    return free_course_angle;
}


static
double
get_free_forward_width( const rcsc::WorldModel & wm,
                        const rcsc::Vector2D & point )
{
    const int VALID_PLAYER_ACCURACY = 20;

    rcsc::AbstractPlayerCont opp_set
        = wm.getPlayerCont( new rcsc::AndPlayerPredicate
                            ( new rcsc::OpponentOrUnknownPlayerPredicate( wm ),
                              new rcsc::CoordinateAccuratePlayerPredicate
                              ( VALID_PLAYER_ACCURACY ),
                              new rcsc::XCoordinateBackwardPlayerPredicate
                              ( wm.offsideLineX() + 0.01 ),
                              new rcsc::XCoordinateForwardPlayerPredicate( point.x ) ) );

    double free_width = std::min( 20.0,
                                  rcsc::AbstractPlayerObject::get_minimum_evaluation
                                  ( opp_set,
                                    new rcsc::AbsYDiffPlayerEvaluator( point ) ) );

    return free_width;
}


#if 0
static
int
get_pass_cut_count( const rcsc::WorldModel & wm,
                    const rcsc::Vector2D & point,
                    const rcsc::Vector2D & ball_pos )
{
    const int VALID_PLAYER_ACCURACY = 40;

    rcsc::AbstractPlayerCont opp_set
        = wm.getPlayerCont( new rcsc::AndPlayerPredicate
                            ( new rcsc::OpponentOrUnknownPlayerPredicate( wm ),
                              new rcsc::CoordinateAccuratePlayerPredicate
                              ( VALID_PLAYER_ACCURACY ),
                              new rcsc::PointFarPlayerPredicate
                              ( ball_pos, (point - ball_pos).r() ),
                              new rcsc::PointNearPlayerPredicate
                              ( ball_pos, 45.0 ),
                              new rcsc::AbsAngleDiffLessPlayerPredicate
                              ( ball_pos, (point - ball_pos).th(), 14.5 ) ) );

    return opp_set.size();
}
#endif


#if 0
static
int
get_mark_count( const rcsc::WorldModel & wm,
                const rcsc::Vector2D & point,
                const rcsc::Vector2D & ball_pos )
{
    const int VALID_PLAYER_ACCURACY = 20;

    rcsc::AbstractPlayerCont opp_set;

    opp_set = wm.getPlayerCont( new rcsc::AndPlayerPredicate
                                ( new rcsc::OpponentOrUnknownPlayerPredicate( wm ),
                                  new rcsc::CoordinateAccuratePlayerPredicate
                                  ( VALID_PLAYER_ACCURACY ),
                                  new rcsc::XCoordinateForwardPlayerPredicate
                                  ( point.x ),
                                  new rcsc::XCoordinateBackwardPlayerPredicate
                                  ( point.x + 2.0 ),
                                  new rcsc::XCoordinateBackwardPlayerPredicate
                                  ( ball_pos.x + 10.0 ),
                                  new rcsc::YCoordinatePlusPlayerPredicate
                                  ( point.y - 1.0 ),
                                  new rcsc::YCoordinateMinusPlayerPredicate
                                  ( point.y + 1.0 ) ) );

    return opp_set.size();
}
#endif


static
double getDistPlayerNearestToPoint( const rcsc::WorldModel &,
                                    const rcsc::Vector2D & point,
                                    const rcsc::PlayerCont & cont,
                                    int validOpponentThreshold )
{
    static const double MAX_DIST = 65535.0;

    double min_dist2 = MAX_DIST;

    const rcsc::PlayerCont::const_iterator end = cont.end();
    for ( rcsc::PlayerCont::const_iterator it = cont.begin();
          it != end;
          ++it )
    {
        if ( (*it).isGhost() )
        {
            continue;
        }

        if ( validOpponentThreshold != -1 )
        {
            if ( (*it).posCount() > validOpponentThreshold )
            {
                continue;
            }
        }

        double dist2 = ( (*it).pos() - point ).r2();

        if ( dist2 < min_dist2 )
        {
            min_dist2 = dist2;
        }
    }

    return std::sqrt( min_dist2 );
}

static
double getDistOpponentNearestToPoint( const rcsc::WorldModel & wm,
                                      const rcsc::Vector2D & point,
                                      int validOpponentThreshold )
{
    return( getDistPlayerNearestToPoint( wm, point, wm.opponents(),
                                         validOpponentThreshold ) );
}
